class EzUtils:
    def __init__(self):
        self.__version__ = "0.0.15"
    
    def hello_world(self):
        print("hello world!!")

    def get_q1(self,x):
        return x.quantile(.25)

    def get_q3(self,x):
        return x.quantile(.75)


import pandas as pd
import numpy as np
from datetime import datetime
import time

from .EzUtils import EzUtils

class EzBasketEvolving(EzUtils):
    def __init__(self):
        super().__init__()
        self.data = None
        self.customerId = 'customerId'
        self.orderId = 'orderId'
        self.orderDate = 'orderDate'
        self.salesPrice = 'salesPrice'

    @property
    def clear_cach(self):
        self.data = None

    @property
    def echo(self):
        print("BasketEvolving")

    def prep_evolving(self, lags):
        evol_basket = self.data.copy()

        for lag in range(lags+1):
            column = '_{}'.format(lag)
            evol_basket[column] = evol_basket.groupby(self.customerId)[self.salesPrice].shift(lag).round(2)

        evol_basket = evol_basket.groupby(self.customerId).agg(
            {"_{}".format(i): lambda x: x.iloc[-1] for i in range(lags+1)}
        )

        evol_basket.columns = ['lag{}'.format(col) for col in evol_basket.columns]

        return evol_basket.reset_index().fillna(0)

    def prep_evolving_pct(self, lags):
        evol_basket = self.data.copy()

        for lag in range(lags+1):
            column = '_{}'.format(lag)
            evol_basket[column] = evol_basket.groupby(self.customerId)[self.salesPrice].pct_change(lag+1).round(2)

        evol_basket = evol_basket.groupby(self.customerId).agg(
            {"_{}".format(i): lambda x: x.iloc[-1] for i in range(lags+1)}
        )

        evol_basket.columns = ['pct{}'.format(col) for col in evol_basket.columns]

        return evol_basket.reset_index().fillna(0)

    def fit(self, data, lags=2):

        start = time.time()
        self.data = data
        evol_df = self.prep_evolving(lags=lags)
        evol_df_pct = self.prep_evolving_pct(lags=lags)

        final_df = evol_df.merge(evol_df_pct,how='left',left_on=self.customerId,right_on=self.customerId)
        self.clear_cach
        end = time.time()
        print('successfully preped BasketEvolving {:.4f} ms'.format(time.time()-start))      

        return final_df


# from folder.file import class
from .EzUtils import *
from .DataModel import *
from .EzCheck import *
from .EzCohort import *
from .EzCustomerMovement import *
from .EzRFM import *
from .EzBasketEvolving import *
from .EzLog import *
from .EzEvaluation import *
from .EzBenchMark import *
from .EzPipeline import *


import pandas as pd
import numpy as np

from .EzCheck import EzCheck

class EzLog(EzCheck):
    def __init__(self):
        super().__init__()
        self.log = pd.DataFrame()
        self.id = 1
        self.idx = 0
        self.session = 1

    def get_id(self, method):
        if method == 'add':
            self.id += 1

    def get_session(self, method):
        if method == 'add':
            self.session += 1

    def show(self):
        return self.log

    def detail(self, id):
        pass

    def get_single_line(self, data, remark, method, id=None):
        row, col = data.shape
        checked_data = self.check(data)
        data_dict = checked_data.to_dict()
        
        columns = checked_data.index.to_list()
        
        idx = self.idx

        if method == 'add':
            id = self.id
        elif method == 'update':
            id = int(id)
        single_line_df = pd.DataFrame(
            {
                'id':id,
                'remark': remark,
                'n_row':row,
                'n_col':col,
                'features': [columns],
                'data':[data.to_dict()]
            },
            index=[idx]
        )
        for k, v in data_dict.items():
            single_line_df[k] = pd.DataFrame({
                k:[v]
            },index=[idx])

        return single_line_df

    def add(self, data, remark="nope"):        
        single_line_df = self.get_single_line(data,remark, method='add')
        self.get_id('add')

        self.log = pd.concat([self.log,single_line_df],axis=0, ignore_index=True)
        return self.log

    def remove_id(self,id):
        if self.log.shape[0] == 1:
            self.log = pd.DataFrame()
        elif self.log.shape[0] > 1:
            self.log = self.log.loc[self.log['id'] != id]
            self.log.index = np.arange(self.log.shape[0])

        return self.log

    def update(self,data, remark="nope", id=None):
        single_line_df = self.get_single_line(data, remark, method='update', id=id)
        self.log.loc[self.log['id']==id] = single_line_df.values
        
        return self.log

    def get_detail(self, id):
        data = self.log.copy()
        filter = ['id', 'data','remark','n_row','n_col','features']
        data = data.loc[data['id']==id]
        row = data['n_row'].values[0]
        col = data['n_col'].values[0]
        remark = data['remark'].values[0]
        init = pd.DataFrame(index=data['features'].values[0])
        data = data.loc[:, ~data.columns.isin(filter)]

        for column in data.columns:
            init[column] = pd.Series(data[column].values[0])

        print(f"""remark: {remark}\nn_row, n_col: {row}, {col}""")

        return init

    def get_data(self, id):
        kind_df = pd.DataFrame(self.log.loc[self.log['id']==id, 'data'].values[0])
        return kind_df


import pandas as pd
import numpy as np
from datetime import datetime
import time

from .EzUtils import EzUtils

class EzRFM(EzUtils):
    def __init__(self):
        super().__init__()
        self.data = None
        self.date_df = None
        self.snapshot = None
        self.customerId = 'customerId'
        self.orderId = 'orderId'
        self.orderDate = 'orderDate'
        self.salesPrice = 'salesPrice'

    @property
    def clear_cach(self):
        self.data = None
        self.date_df = None
        self.snapshot = None

    @property
    def echo(self):
        print('EzRFM')

    # def get_q1(self,x):
    #     return x.quantile(.25)

    # def get_q3(self,x):
    #     return x.quantile(.75)

    def get_range(self,x):
        return x.max()-x.min()

    def get_recency(self,x):
        return (self.snapshot-x.max()).days

    def get_tenure(self,x):
        return (self.snapshot-x.min()).days

    def prep_rfmt(self, data):

        data = data.groupby(self.customerId).agg({
            self.orderDate:[self.get_recency, self.get_tenure],
            self.orderId:'nunique',
            self.salesPrice:['sum','min',self.get_q1,'mean','median',
                             self.get_q3,'max','std',self.get_range,'skew']
        }).reset_index()

        names = ['Recency','Tenure','Frequency', 'Monetary',
                 'MonetaryMin','MonetaryQ1','MonetaryMean','MonetaryMedian',
                 'MonetaryQ3','MonetaryMax','MonetaryStd','MonetaryRange','MonetarySkew']
        data.columns = [self.customerId] + ['total' +  name for name in names]
        
        column_display = ['Recency','Frequency', 'Monetary', 'Tenure',
                 'MonetaryMin','MonetaryQ1','MonetaryMean','MonetaryMedian',
                 'MonetaryQ3','MonetaryMax','MonetaryStd','MonetaryRange','MonetarySkew']
        data = data[[self.customerId] + ['total' +  col for col in column_display]]

        return data.fillna(0).round(2)

    def prep_rfmt_date(self, data, column, method):
        data = data.groupby([self.customerId, column]).agg({
            self.orderId:'nunique',
            self.salesPrice:'sum'
        }).reset_index()

        data = data.groupby(self.customerId).agg({
            column:'nunique',
            self.orderId:['mean','median','std'],
            self.salesPrice:['min', self.get_q1, 'mean', 'median', 
                             self.get_q3, 'max','std', self.get_range,'skew']
        }).reset_index()

        names = ['FrequencySum','FrequencyMean','FrequencyMedian','FrequencyStd',
                 'MonetaryMin','MonetaryQ1','MonetaryMean','MonetaryMedian',
                 'MonetaryQ3','MonetaryMax','MonetaryStd','MonetaryRange','MonetarySkew']
        data.columns = [self.customerId] + [method +  name for name in names]

        denominator = {'year': 365, 'quarter':90, 'month': 30, 'week': 7}

        data.insert(loc=1, column=f'{method}Tenure', value=self.date_df['totalTenure'] / denominator[method])
        data.insert(loc=1, column=f'{method}Recency', value=self.date_df['totalRecency'] / denominator[method])

        return data

    def fit(self, data, method=None, snapshot=None):
        
        start = time.time()

        if snapshot == None:
            self.snapshot = data[self.orderDate].max() + np.timedelta64(1,'D')
        else:
            self.snapshot = pd.datetime(snapshot)
        
        rfmt_df = self.prep_rfmt(data)
        self.date_df = rfmt_df[['totalRecency','totalTenure']]
        method_dict = {'year':'year','yearQuarter':'quarter',
                       'yearMonth':'month','yearWeek':'week'}

        if method == None:
            rfmt_df = rfmt_df
        else:
            if method == 'year':
                columns = ['year']
            elif method == 'quarter':
                columns = ['yearQuarter']
            elif method == 'month':
                columns = ['yearMonth']
            elif method == 'week':
                columns = ['yearWeek']
            elif method == 'full':
                columns = ['year','yearQuarter','yearMonth','yearWeek']

            for enum, column in enumerate(columns):        
                rfmt_date = self.prep_rfmt_date(data, column, method_dict[column])
                rfmt_df = pd.concat([rfmt_df, rfmt_date.drop(self.customerId, axis=1)],axis=1)
        self.clear_cach
        print('successfully preped EzRFM {:.4f} ms'.format(time.time()-start))

        return rfmt_df.round(4)


import pandas as pd
import numpy as np
# from sklearn.linear_model import LogisticRegression
# from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
# from sklearn.svm import SVC
# from xgboost import XGBClassifier
# from sklearn.tree import DecisionTreeClassifier
# from sklearn.dummy import DummyClassifier

from sklearn.linear_model import *
from sklearn.ensemble import *
from sklearn.svm import *
from xgboost import *
from sklearn.tree import *
from sklearn.dummy import *

from .EzEvaluation import EzEvaluation

class EzBenchMark(EzEvaluation): 
    def __init__(self):
        super().__init__()
        self.artifact_ = {}

    def get_single_line(self, X_train, y_train, X_test, y_test, remark, 
                        model_name, performance, artifact, method, id=None):
        row, col = X_train.shape
        
        columns = X_train.columns.to_list()
        
        idx = self.idx

        if method == 'add':
            id = self.id
            session = self.session
        elif method == 'update':
            id = int(id)
        single_line_df = pd.DataFrame(
            {
                'id':id,
                'session':session,
                'remark': remark,
                'model': model_name,
                'performance': [performance.to_dict()],
                'n_row':row,
                'n_col':col,
                'features': [columns],
                'X_train': [X_train.to_dict()],
                'y_train': [y_train.to_dict()],
                'X_test': [X_test.to_dict()],
                'y_test': [y_test.to_dict()],

            },
            index=[idx]
        )
        self.artifact_[id] = artifact

        return single_line_df

    def get_model(self, id):
        return self.artifact_[id]

    def add(self, X_train, y_train, X_test, y_test,
            model_name, performance, artifact, remark="nope"):        

        single_line_df = self.get_single_line(X_train, y_train, 
                                              X_test, y_test, remark=remark,
                                              model_name=model_name, performance=performance, 
                                              artifact=artifact, method='add', id=None)
        self.get_id('add')

        self.log = pd.concat([self.log,single_line_df],axis=0, ignore_index=True)
        return self.log

    def remove_session(self,session):
        if self.log.shape[0] == 1:
            self.log = pd.DataFrame()
        elif self.log.shape[0] > 1:
            self.log = self.log.loc[self.log['session'] != session]
            self.log.index = np.arange(self.log.shape[0])

        return self.log

    # *design logic first*
    # def update(self,data, remark="nope", id=None):
    #     single_line_df = self.get_single_line(data, remark, method='update', id=id)
    #     self.log.loc[self.log['id']==id] = single_line_df.values
        
    #     return self.log

    def train_model(self, X_train, y_train, X_test, y_test, model_dict, probability='target'):

        model_results = {}
        model_artifacts = {}
        for model_name, base_model in model_dict.items():

            print(f"start {model_name}")

            model = base_model
            model.fit(X_train,y_train)
            model_artifacts[model_name] = model
            probabilities = model.predict_proba(X_test)
            # if probability=='target':
            #     model_results[model_name] = probabilities[:,1]
            # elif probability=='all':
            #     model_results[model_name] = probabilities[:,:]

        # return model_artifacts, model_results
        return model_artifacts

    def benchmark(self, X_train, y_train, X_test, y_test, model_dict, remark, kind):
        
        # model_artifacts, model_results  = self.train_model(X_train, y_train, X_test, y_test, model_dict)
        model_artifacts = self.train_model(X_train, y_train, X_test, y_test, model_dict)
        
        for model_name, artifact in model_artifacts.items():

            if kind == 'classification':
                performance = self.get_classification_report(model_name,y_test,artifact.predict_proba(X_test)[:,1],cost=0,benefit=0)
            elif kind == 'regression':
                performance = self.get_regression_report(model_name,y_test,artifact.predict(X_test))

            self.add(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test,
                                 model_name=model_name, performance=performance, artifact=artifact, remark=remark)
            
        self.get_session(method='add')


    def classification_benchmark(self, X_train, y_train, X_test, y_test, model_dict, remark='nope'):
        self.benchmark(X_train, y_train, X_test, y_test, model_dict, remark=remark, kind='classification')
        return self.log

    def regression_benchmark(self, X_train, y_train, X_test, y_test, model_dict, remark='nope'):    
        self.benchmark(X_train, y_train, X_test, y_test, model_dict, remark=remark, kind='regression')
        return self.log

    def get_data(self, id, kind='X_test'):
        if kind in ['X_train','X_test']:
            kind_df = pd.DataFrame(self.log.loc[self.log['id']==id, kind].values[0])
        elif kind in ['y_train','y_test']:
            kind_df = pd.Series(self.log.loc[self.log['id']==id, kind].values[0])
        return kind_df

    def get_performance(self, id):
        kind_df = pd.DataFrame(self.log.loc[self.log['id']==id, 'performance'].values[0])
        return kind_df

    def model_performance(self, id=None, session=None):
        
        if type(id) == int:
            id = [id]
        if type(session) == int:
            session = [session]

        if (id==None) & (session==None):
            performance_df = self.log.loc[:, ['id','session','performance']]
        elif (id==None) & (session!=None):
            performance_df = self.log.loc[self.log['session'].isin(session), ['id','session','performance']]
        elif (id!=None) & (session==None):
            performance_df = self.log.loc[self.log['id'].isin(id), ['id','session','performance']]
        else:
            performance_df = self.log.loc[(self.log['id'].isin(id)) & (self.log['session'].isin(session)), 
                                          ['id','session','performance']]

        ids = performance_df['id'].unique()
        final_df = pd.DataFrame()
        for id in ids:
            final_df = pd.concat([final_df,
                                  self.get_performance(id=id)],
                                 axis=0, ignore_index=True)
        return final_df